@@ -1,74 +1,25 @@
 # -*- coding: utf-8 -*-
 
-from collections import defaultdict
+from __future__ import absolute_import
 
 from django.core.cache.backends.base import BaseCache, InvalidCacheBackendError
 from django.core.exceptions import ImproperlyConfigured
-from django.utils import importlib
 from django.utils.encoding import smart_unicode, smart_str
 from django.utils.datastructures import SortedDict
+from django.utils import importlib
 
 try:
     import cPickle as pickle
 except ImportError:
     import pickle
 
-try:
-    import redis
-except ImportError:
-    raise InvalidCacheBackendError("Redis cache backend requires the 'redis-py' library")
+from collections import defaultdict
 
-from redis.connection import UnixDomainSocketConnection, Connection
+from redis import Redis
 from redis.connection import DefaultParser
+from .util import CacheKey, ConnectionPoolHandler
 
-class CacheKey(object):
-    """
-    A stub string class that we can use to check if a key was created already.
-    """
-    def __init__(self, key):
-        self._key = key
-
-    def __eq__(self, other):
-        return self._key == other
-
-    def __str__(self):
-        return self.__unicode__()
-
-    def __repr__(self):
-        return self.__unicode__()
-
-    def __unicode__(self):
-        return smart_str(self._key)
-
-
-class CacheConnectionPool(object):
-    _connection_pool = None
-
-    def get_connection_pool(self, host='127.0.0.1', port=6379, db=1,
-                password=None, parser_class=None, unix_socket_path=None):
-
-        if self._connection_pool is None:
-            connection_class = unix_socket_path \
-                and UnixDomainSocketConnection or Connection
-
-            kwargs = {
-                'db': db,
-                'password': password,
-                'connection_class': connection_class,
-                'parser_class': parser_class,
-            }
-
-            if unix_socket_path is None:
-                kwargs.update({'host': host, 'port': port})
-            else:
-                kwargs['path'] = unix_socket_path
-
-            self._connection_pool = redis.ConnectionPool(**kwargs)
-
-        return self._connection_pool
-
-# ConnectionPools keyed off the connection parameters
-pools = defaultdict(CacheConnectionPool)
+import re
 
 class RedisCache(BaseCache):
     _pickle_version = -1
@@ -80,15 +31,10 @@ class RedisCache(BaseCache):
         self._init(server, params)
         super(RedisCache, self).__init__(params)
 
-    def _init(self, server, params):
-        super(RedisCache, self).__init__(params)
-        self._server = server
-        self._params = params
-        self._options = params.get('OPTIONS', {})
-
+    def _connect(self):
         unix_socket_path = None
-        if ':' in self.server:
-            host, port = self.server.split(':')
+        if ':' in self._server:
+            host, port = self._server.split(':')
             try:
                 port = int(port)
             except (ValueError, TypeError):
@@ -96,74 +42,74 @@ class RedisCache(BaseCache):
 
         else:
             host, port = None, None
-            unix_socket_path = self.server
-        
-        if "PICKLE_VERSION" in self._options:
-            try:
-                self._pickle_version = int(self._options['PICKLE_VERSION'])
-            except (ValueError, TypeError):
-                raise ImproperlyConfigured("PICKLE_VERSION value must be an integer")
+            unix_socket_path = self._server
 
+        # parse database
+        _db = self._params.get('db', self._options.get('DB', 1))
+        try:
+            _db = int(_db)
+        except (ValueError, TypeError):
+            raise ImproperlyConfigured("db value must be an integer")
+
+        # params for connection pool.
         kwargs = {
-            'db': self.db,
+            'db': _db,
             'password': self.password,
             'host': host,
             'port': port,
             'unix_socket_path': unix_socket_path,
         }
 
-        pool_key = ':'.join([str(v) for v in kwargs.values()])
+        connection_pool = ConnectionPoolHandler()\
+            .connection_pool(parser_class=self.parser_class, **kwargs)
+        self._client = Redis(connection_pool=connection_pool)
 
-        global pools
-        connection_pool = pools[pool_key].get_connection_pool(
-            parser_class=self.parser_class, **kwargs)
+    def _init(self, server, params):
+        super(RedisCache, self).__init__(params)
+        self._server = server
+        self._params = params
+        self._options = params.get('OPTIONS', {})
 
-        self._client = redis.Redis(connection_pool=connection_pool)
+        if "PICKLE_VERSION" in self._options:
+            try:
+                self._pickle_version = int(self._options['PICKLE_VERSION'])
+            except (ValueError, TypeError):
+                raise ImproperlyConfigured("PICKLE_VERSION value must be an integer")
+
+        self._connect()
 
     def make_key(self, key, version=None):
         if not isinstance(key, CacheKey):
             key = CacheKey(super(RedisCache, self).make_key(key, version))
         return key
 
-    def incr_version(self, key, delta=1, version=None):
+    def incr_version(self, key, delta=1, version=None, client=None):
         """
         Adds delta to the cache version for the supplied key. Returns the
         new version.
-
-        Note: In Redis 2.0 you cannot rename a volitle key, so we have to move
-        the value from the old key to the new key and maintain the ttl.
         """
+        if client is None:
+            client = self._client
+
         if version is None:
             version = self.version
+
         old_key = self.make_key(key, version)
-        value = self.get(old_key, version=version)
-        ttl = self._client.ttl(old_key)
+        value = self.get(old_key, version=version, client=client)
+        ttl = client.ttl(old_key)
+
         if value is None:
             raise ValueError("Key '%s' not found" % key)
-        new_key = self.make_key(key, version=version+delta)
-        # TODO: See if we can check the version of Redis, since 2.2 will be able
-        # to rename volitile keys.
+
+        if isinstance(key, CacheKey):
+            new_key = self.make_key(key.original_key(), version=version+delta)
+        else:
+            new_key = self.make_key(key, version=version+delta)
+
         self.set(new_key, value, timeout=ttl)
-        self.delete(old_key)
+        self.delete(old_key, client=client)
         return version + delta
 
-    @property
-    def server(self):
-        return self._server or "127.0.0.1:6379"
-
-    @property
-    def params(self):
-        return self._params or {}
-
-    @property
-    def db(self):
-        _db = self.params.get('db', self._options.get('DB', 1))
-        try:
-            _db = int(_db)
-        except (ValueError, TypeError):
-            raise ImproperlyConfigured("db value must be an integer")
-        return _db
-
     @property
     def password(self):
         return self._options.get('PASSWORD', None)
@@ -173,6 +119,7 @@ class RedisCache(BaseCache):
         cls = self._options.get('PARSER_CLASS', None)
         if cls is None:
             return DefaultParser
+
         mod_path, cls_name = cls.rsplit('.', 1)
         try:
             mod = importlib.import_module(mod_path)
@@ -193,29 +140,39 @@ class RedisCache(BaseCache):
         for c in self._client.connection_pool._available_connections:
             c.disconnect()
 
-    def add(self, key, value, timeout=None, version=None):
+    def add(self, key, value, timeout=None, version=None, client=None):
         """
         Add a value to the cache, failing if the key already exists.
 
         Returns ``True`` if the object was added, ``False`` if not.
         """
-        key = self.make_key(key, version=version)
-        if self._client.exists(key):
-            return False
-        return self.set(key, value, timeout)
 
-    def get(self, key, default=None, version=None):
+        if client is None:
+            client = self._client
+
+        key = self.make_key(key, version=version)
+
+        if client.exists(key):
+            return False
+
+        return self.set(key, value, timeout, client=client)
+
+    def get(self, key, default=None, version=None, client=None):
         """
         Retrieve a value from the cache.
 
         Returns unpickled value if key is found, the default if not.
         """
+        if client is None:
+            client = self._client
+
         key = self.make_key(key, version=version)
-        value = self._client.get(key)
+        value = client.get(key)
+
         if value is None:
             return default
-        return self.unpickle(value)
 
+        return self.unpickle(value)
 
     def _set(self, key, value, timeout, client):
         if timeout == 0:
@@ -229,21 +186,37 @@ class RedisCache(BaseCache):
         """
         Persist a value to the cache, and set an optional expiration time.
         """
+
         if not client:
             client = self._client
 
         key = self.make_key(key, version=version)
         if timeout is None:
             timeout = self.default_timeout
-    
+
         result = self._set(key, self.pickle(value), int(timeout), client)
         return result
 
-    def delete(self, key, version=None):
+    def delete(self, key, version=None, client=None):
         """
         Remove a key from the cache.
         """
-        self._client.delete(self.make_key(key, version=version))
+        if client is None:
+            client = self._client
+
+        client.delete(self.make_key(key, version=version))
+
+    def delete_pattern(self, pattern, version=None, client=None):
+        """
+        Remove all keys matching pattern.
+        """
+
+        if client is None:
+            client = self._client
+
+        pattern = self.make_key(pattern, version=version)
+        keys = client.keys(pattern)
+        client.delete(*keys)
 
     def delete_many(self, keys, version=None):
         """
@@ -279,19 +252,16 @@ class RedisCache(BaseCache):
         if not keys:
             return {}
         recovered_data = SortedDict()
+
         new_keys = map(lambda key: self.make_key(key, version=version), keys)
         map_keys = dict(zip(new_keys, keys))
+
         results = self._client.mget(new_keys)
         for key, value in zip(new_keys, results):
             if value is None:
                 continue
-            try:
-                value = int(value)
-            except (ValueError, TypeError):
-                value = self.unpickle(value)
-            if isinstance(value, basestring):
-                value = smart_unicode(value)
-            recovered_data[map_keys[key]] = value
+
+            recovered_data[map_keys[key]] = self.unpickle(value)
         return recovered_data
 
     def set_many(self, data, timeout=None, version=None):
@@ -307,35 +277,40 @@ class RedisCache(BaseCache):
             self.set(key, value, timeout, version=version, client=pipeline)
         pipeline.execute()
 
-    def incr(self, key, delta=1, version=None):
+    def incr(self, key, delta=1, version=None, client=None):
         """
         Add delta to value in the cache. If the key does not exist, raise a
         ValueError exception.
         """
-        key = self.make_key(key, version=version)
-        exists = self._client.exists(key)
 
-        if not exists:
+        if client is None:
+            client = self._client
+
+        key = self.make_key(key, version=version)
+        exists = client.exists(key)
+
+        if not client.exists(key):
             raise ValueError("Key '%s' not found" % key)
 
-        value = self.get(key) + delta
-        self.set(key, value)
+        value = self.get(key, version=version, client=client) + delta
+        self.set(key, value, version=version, client=client)
         return value
 
-    def decr(self, key, delta=1, version=None):
+    def decr(self, key, delta=1, version=None, client=None):
         """
         Decreace delta to value in the cache. If the key does not exist, raise a
         ValueError exception.
         """
+        if client is None:
+            client = self._client
 
         key = self.make_key(key, version=version)
-        exists = self._client.exists(key)
 
-        if not exists:
+        if not client.exists(key):
             raise ValueError("Key '%s' not found" % key)
 
-        value = self.get(key) - delta
-        self.set(key, value)
+        value = self.get(key, version=version, client=client) - delta
+        self.set(key, value, version=version, client=client)
         return value
 
     def has_key(self, key, version=None):
@@ -348,4 +323,153 @@ class RedisCache(BaseCache):
 
     # Other not default and not standar methods.
     def keys(self, search):
-        return list(set(map(lambda x: x.split(":", 2)[2], self._client.keys(search))))
+        pattern = self.make_key(search)
+        return list(set(map(lambda x: x.split(":", 2)[2], self._client.keys(pattern))))
+
+
+from .hash_ring import HashRing
+_findhash = re.compile('.*\{(.*)\}.*', re.I)
+
+class ShardedRedisCache(RedisCache):
+    connections = {}
+    nodes = []
+
+    def _connect(self):
+        if not isinstance(self._server, (tuple, list)):
+            raise ImproperlyConfigured("LOCATION must be a list or tuple")
+
+        for location in self._server:
+            try:
+                host, port, db = location.split(":")
+            except ValueError:
+                try:
+                    host, port = location.split(":")
+                    db = 1
+                except ValueError:
+                    raise ImproperlyConfigured("invalid location string, this must be <host>:<port>:<db>")
+
+            params = {
+                'db': db,
+                'password': self.password,
+            }
+            if host == "unix":
+                params['unix_socket_path'] = port
+            else:
+                params['unix_socket_path'] = None
+                params['host'], params['port'] = host, int(port)
+
+            connection_pool = ConnectionPoolHandler()\
+                .connection_pool(parser_class=self.parser_class, **params)
+
+            self.connections[location] = Redis(connection_pool=connection_pool)
+            self.nodes.append(location)
+
+        self.ring = HashRing(self.nodes)
+
+    def get_server_name(self, _key):
+        key = str(_key)
+        g = _findhash.match(key)
+        if g != None and len(g.groups()) > 0:
+            key = g.groups()[0]
+        name = self.ring.get_node(key)
+        return name
+
+    def get_server(self,key):
+        name = self.get_server_name(key)
+        return self.connections[name]
+
+    def add(self,  key, value, timeout=None, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardedRedisCache, self).add(key=key, value=value,
+                                        version=version, client=client)
+
+    def get(self,  key, default=None, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+        return super(ShardedRedisCache, self).get(key=key, default=default,
+                                                version=version, client=client)
+
+    def get_many(self, keys, version=None):
+        if not keys:
+            return {}
+
+        recovered_data = SortedDict()
+        new_keys = map(lambda key: self.make_key(key, version=version), keys)
+        map_keys = dict(zip(new_keys, keys))
+
+        for key in new_keys:
+            client = self.get_server(key)
+            value = self.get(key=key, version=version, client=client)
+
+            if value is None:
+                continue
+
+            recovered_data[map_keys[key]] = value
+        return recovered_data
+
+    def set(self, key, value, timeout=None, version=None, client=None):
+        """
+        Persist a value to the cache, and set an optional expiration time.
+        """
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardedRedisCache, self).set(key=key, value=value, timeout=timeout,
+                                                        version=version, client=client)
+
+
+    def set_many(self, data, timeout=None, version=None):
+        """
+        Set a bunch of values in the cache at once from a dict of key/value
+        pairs. This is much more efficient than calling set() multiple times.
+
+        If timeout is given, that timeout will be used for the key; otherwise
+        the default cache timeout will be used.
+        """
+        for key, value in data.iteritems():
+            self.set(key, value, timeout, version=version)
+
+    def delete(self, key, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardedRedisCache, self).delete(key=key, version=version, client=client)
+
+    def delete_many(self, keys, version=None):
+        """
+        Remove multiple keys at once.
+        """
+        for key in map(lambda key: self.make_key(key, version=version), keys):
+            client = self.get_server(key)
+            self.delete(key, client=client)
+
+    def incr_version(self, key, delta=1, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardedRedisCache, self).incr_version(key=key, delta=delta,
+                                                    version=version, client=client)
+
+    def incr(self, key, delta=1, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardedRedisCache, self).incr(key=key, delta=delta,
+                                        version=version, client=client)
+
+
+    def decr(self, key, delta=1, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super(ShardedRedisCache, self).decr(key=key, delta=delta,
+                                        version=version, client=client)
